import datetime
from collections import UserDict

from typing import Dict, Any

from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.log import LoggerAdapter, default_log
from golem.core.optimisers.dynamic_graph_requirements import DynamicGraphRequirements
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.utilities.utilities import determine_n_jobs


class ApiParams(UserDict):
    """
    Class to further distribute params specified params in API between the following classes:
    `GraphRequirements`, `GraphGenerationParams`, `GPAlgorithmParameters`.
    """
    def __init__(self, input_params: Dict[str, Any], n_jobs: int = -1, timeout: float = 5):
        self.log: LoggerAdapter = default_log(self)
        self.n_jobs: int = determine_n_jobs(n_jobs)
        self.timeout = timeout

        self._input_params = input_params
        self._input_params['timeout'] = timeout if isinstance(timeout, datetime.timedelta) else datetime.timedelta(minutes=timeout)
        self._default_common_params = self.get_default_common_params()
        super().__init__(self._input_params)

    def get_default_common_params(self):
        """ Common params that do not belong to any category
        (from `GPAlgorithmParameters`, `GraphGenerationParams`, `GraphRequirements`). """
        default_common_params = {
            'optimizer': EvoGraphOptimizer,
            'initial_graphs': list(),
            'objective': None
        }
        self.log.info("EvoGraphOptimizer was used as default optimizer, "
                      "will be overwritten by specified one if there is any.")
        return default_common_params

    def get_default_graph_generation_params(self):
        """ Default graph generations params to minimize the number of arguments that must be specified in API.
        Need to be hardcoded like that since the list of input arguments is not the same as the class fields list. """
        default_graph_generation_params = {
            'adapter': BaseNetworkxAdapter(),
            'rules_for_constraint': DEFAULT_DAG_RULES,
            'advisor': None,
            'node_factory': None,
            'random_graph_factory': None,
            'available_node_types': None,
            'remote_evaluator': None
        }
        self.log.info("BaseNetworkxAdapter was used as default adapter, "
                      "will be overwritten by specified one if there is any.")
        return default_graph_generation_params

    def get_gp_algorithm_parameters(self) -> GPAlgorithmParameters:
        default_gp_algorithm_params_dict = dict(list(vars(GPAlgorithmParameters()).items()))
        k_pop = []
        for k, v in self._input_params.items():
            if k in default_gp_algorithm_params_dict:
                default_gp_algorithm_params_dict[k] = self._input_params[k]
                k_pop.append(k)
        for k in k_pop:
            self._input_params.pop(k)
        return GPAlgorithmParameters(**default_gp_algorithm_params_dict)

    def get_graph_generation_parameters(self) -> GraphGenerationParams:
        default_graph_generation_params_dict = self.get_default_graph_generation_params()
        k_pop = []
        for k, v in self._input_params.items():
            if k in default_graph_generation_params_dict:
                default_graph_generation_params_dict[k] = self._input_params[k]
                k_pop.append(k)
        for k in k_pop:
            self._input_params.pop(k)
        ggp = GraphGenerationParams(**default_graph_generation_params_dict)
        return ggp

    def get_graph_requirements(self) -> GraphRequirements:
        default_graph_requirements_params_dict = dict(list(vars(GraphRequirements()).items()))
        # if there are any custom domain specific graph requirements params
        is_custom_graph_requirements_params = \
            any([k not in default_graph_requirements_params_dict for k in self._input_params])
        for k, v in self._input_params.items():
            # add all parameters except common left unused after GPAlgorithmParameters and GraphGenerationParams
            # initialization, since it can be custom domain specific params
            if k not in self._default_common_params:
                default_graph_requirements_params_dict[k] = self._input_params[k]
        if is_custom_graph_requirements_params:
            return DynamicGraphRequirements(default_graph_requirements_params_dict)
        else:
            return GraphRequirements(**default_graph_requirements_params_dict)

    def get_actual_common_params(self) -> Dict[str, Any]:
        for k, v in self._input_params.items():
            if k in self._default_common_params:
                self._default_common_params[k] = v
        return self._default_common_params



from functools import partial
from typing import Tuple, Optional

from hyperopt import Trials, fmin, space_eval

from golem.core.constants import MIN_TIME_FOR_TUNING_IN_SEC
from golem.core.optimisers.graph import OptGraph
from golem.core.tuning.hyperopt_tuner import HyperoptTuner, get_node_parameters_for_hyperopt
from golem.core.tuning.tuner_interface import DomainGraphForTune


class SimultaneousTuner(HyperoptTuner):
    """
        Class for hyperparameters optimization for all nodes simultaneously
    """

    def _tune(self, graph: DomainGraphForTune, show_progress: bool = True) -> DomainGraphForTune:
        """ Function for hyperparameters tuning on the entire graph

        Args:
            graph: graph which hyperparameters will be tuned
            show_progress: shows progress of tuning if True

        Returns:
            Graph with tuned hyperparameters
        """
        parameters_dict, init_parameters = self._get_parameters_for_tune(graph)
        remaining_time = self._get_remaining_time()

        if self._check_if_tuning_possible(graph, len(parameters_dict) > 0, remaining_time):
            trials = Trials()

            try:
                # try searching using initial parameters
                # (uses original search space with fixed initial parameters)
                trials, init_trials_num = self._search_near_initial_parameters(
                    partial(self._objective,
                            graph=graph,
                            unchangeable_parameters=init_parameters),
                    parameters_dict,
                    init_parameters,
                    trials,
                    remaining_time,
                    show_progress)
                remaining_time = self._get_remaining_time()
                if remaining_time > MIN_TIME_FOR_TUNING_IN_SEC:
                    fmin(partial(self._objective, graph=graph),
                         parameters_dict,
                         trials=trials,
                         algo=self.algo,
                         max_evals=self.iterations,
                         show_progressbar=show_progress,
                         early_stop_fn=self.early_stop_fn,
                         timeout=remaining_time)
                else:
                    self.log.message('Tunner stopped after initial search due to the lack of time')

                best = space_eval(space=parameters_dict, hp_assignment=trials.argmin)
                # check if best point was obtained using search space with fixed initial parameters
                is_best_trial_with_init_params = trials.best_trial.get('tid') in range(init_trials_num)
                if is_best_trial_with_init_params:
                    best = {**best, **init_parameters}

                final_graph = self.set_arg_graph(graph=graph, parameters=best)

                self.was_tuned = True

            except Exception as ex:
                self.log.warning(f'Exception {ex} occurred during tuning')
                final_graph = graph
        else:
            final_graph = graph
        return final_graph

    def _get_parameters_for_tune(self, graph: OptGraph) -> Tuple[dict, dict]:
        """ Method for defining the search space

        Args:
            graph: graph to be tuned

        Returns:
            parameters_dict: dict with operation names and parameters
            initial_parameters: dict with initial parameters of the graph
        """

        parameters_dict = {}
        initial_parameters = {}
        for node_id, node in enumerate(graph.nodes):
            # Assign unique prefix for each model hyperparameter
            # label - number of node in the graph
            tunable_node_params, initial_node_params = get_node_parameters_for_hyperopt(self.search_space,
                                                                                        node_id=node_id,
                                                                                        node=node)
            parameters_dict.update(tunable_node_params)
            initial_parameters.update(initial_parameters)

        return parameters_dict, initial_parameters

    def _objective(self, parameters_dict: dict, graph: OptGraph, unchangeable_parameters: Optional[dict] = None) \
            -> float:
        """
        Objective function for minimization problem

        Args:
            parameters_dict: dict which contains new graph hyperparameters
            graph: graph to optimize
            unchangeable_parameters: dict with parameters that should not be changed

        Returns:
             metric_value: value of objective function
        """

        # replace new parameters with parameters
        if unchangeable_parameters:
            parameters_dict = {**parameters_dict, **unchangeable_parameters}

        # Set hyperparameters for every node
        graph = self.set_arg_graph(graph, parameters_dict)

        metric_value = self.get_metric_value(graph=graph)

        return metric_value



import os.path
import _pickle as pickle
import random
import re
from functools import partial
from typing import Union, Sequence, Optional, Callable

from mabwiser.mab import MAB, LearningPolicy
from scipy.special import softmax

from golem.core.dag.graph import Graph
from golem.core.dag.graph_node import GraphNode
from golem.core.optimisers.adaptive.operator_agent import OperatorAgent, ActType, ObsType, ExperienceBuffer
from golem.core.optimisers.adaptive.reward_agent import FitnessRateRankRewardTransformer
from golem.core.optimisers.adaptive.utils import get_callable_name
from golem.core.paths import default_data_dir


class MultiArmedBanditAgent(OperatorAgent):
    def __init__(self,
                 actions: Sequence[ActType],
                 n_jobs: int = 1,
                 enable_logging: bool = True,
                 decaying_factor: float = 1.0,
                 path_to_save: Optional[str] = None,
                 is_initial_fit: bool = True):
        super().__init__(actions=actions, enable_logging=enable_logging)
        self.actions = list(actions)
        self._indices = list(range(len(actions)))
        # str because parent operator for mutation is stored as string for custom mutations serialisation
        self._arm_by_action = dict(map(lambda x, y: (get_callable_name(x), y), actions, self._indices))
        self._agent = MAB(arms=self._indices,
                          learning_policy=LearningPolicy.EpsilonGreedy(epsilon=0.4),
                          n_jobs=n_jobs)
        self._reward_agent = FitnessRateRankRewardTransformer(decaying_factor=decaying_factor)
        if is_initial_fit:
            self._initial_fit()
        self._path_to_save = path_to_save

    def _initial_fit(self):
        n = len(self.actions)
        uniform_rewards = [1. / n] * n
        self._agent.fit(decisions=self._indices, rewards=uniform_rewards)

    def choose_action(self, obs: ObsType) -> ActType:
        arm = self._agent.predict()
        action = self.actions[arm]
        return action

    def get_action_values(self, obs: Optional[ObsType] = None) -> Sequence[float]:
        prob_dict = self._agent.predict_expectations()
        prob_list = [prob_dict[i] for i in range(len(prob_dict))]
        return prob_list

    def get_action_probs(self, obs: Optional[ObsType] = None) -> Sequence[float]:
        return softmax(self.get_action_values())

    def choose_nodes(self, graph: Graph, num_nodes: int = 1) -> Union[GraphNode, Sequence[GraphNode]]:
        subject_nodes = random.sample(graph.nodes, k=num_nodes)
        return subject_nodes[0] if num_nodes == 1 else subject_nodes

    def partial_fit(self, experience: ExperienceBuffer):
        """Continues learning of underlying agent with new experience."""
        _, arms, processed_rewards = self._get_experience(experience)
        self._agent.partial_fit(decisions=arms, rewards=processed_rewards)

    def _get_experience(self, experience: ExperienceBuffer):
        """ Get experience from ExperienceBuffer, process rewards and log. """
        obs, actions, rewards = experience.retrieve_experience()
        arms = [self._arm_by_action[action] for action in actions]
        processed_rewards = self._reward_agent.get_rewards_for_arms(rewards, arms)
        self._dbg_log(obs, actions, processed_rewards)
        return obs, arms, processed_rewards

    def save(self, path_to_save: Optional[str] = None):
        """ Saves bandit to specified file. """

        if not path_to_save:
            path_to_save = self._path_to_save

        # if path was not specified at all
        if not path_to_save:
            path_to_save = os.path.join(default_data_dir(), 'MAB')

        if not path_to_save.endswith('.pkl'):
            os.makedirs(path_to_save, exist_ok=True)
            mabs_num = [int(name.split('_')[0]) for name in os.listdir(path_to_save)
                        if re.fullmatch(r'\d_mab.pkl', name)]
            if not mabs_num:
                max_saved_mab = 0
            else:
                max_saved_mab = max(mabs_num) + 1
            path_to_file = os.path.join(path_to_save, f'{max_saved_mab}_mab.pkl')
        else:
            path_to_dir = os.path.dirname(path_to_save)
            os.makedirs(path_to_dir, exist_ok=True)
            path_to_file = path_to_save
        with open(path_to_file, 'wb') as f:
            pickle.dump(self, f)
        self._log.info(f"MAB was saved to {path_to_file}")

    @staticmethod
    def load(path: str):
        """ Loads bandit from the specified file. """
        with open(path, 'rb') as f:
            mab = pickle.load(f)
        return mab



from typing import Sequence, List

from golem.core.optimisers.adaptive.mab_agents.contextual_mab_agent import ContextualMultiArmedBanditAgent
from golem.core.optimisers.adaptive.neural_mab import NeuralMAB
from golem.core.optimisers.adaptive.context_agents import ContextAgentTypeEnum
from golem.core.optimisers.adaptive.common_types import ActType


def get_context_size(context_agent_type: ContextAgentTypeEnum, available_operations: List[str]) -> int:
    """
    Returns the context size based on the context agent type and available operations.
    """
    if context_agent_type == ContextAgentTypeEnum.nodes_num:
        return 1  # returns 1 as nodes_num context represents the number of nodes in the graph
    elif context_agent_type == ContextAgentTypeEnum.operations_quantity:
        return len(available_operations)
    elif context_agent_type == ContextAgentTypeEnum.adjacency_matrix:
        return len(available_operations) ** 2
    elif context_agent_type == ContextAgentTypeEnum.feather_graph:
        return 500  # embedding size for FeatherGraph with default parameters
    elif context_agent_type == ContextAgentTypeEnum.labeled_edges:
        return 100  # fixed context size; can be edited in labeled_edges function
    else:
        raise ValueError(f"Context agent type {context_agent_type} is not supported for NeuralContextualMABAgent.")


class NeuralContextualMultiArmedBanditAgent(ContextualMultiArmedBanditAgent):
    """ Neural Contextual Multi-Armed bandit.
    Observations can be encoded with the use of Neural Networks, but still there are some restrictions
    to guarantee convergence. """

    def __init__(self,
                 actions: Sequence[ActType],
                 context_agent_type: ContextAgentTypeEnum,
                 available_operations: List[str],
                 n_jobs: int = 1,
                 enable_logging: bool = True,
                 decaying_factor: float = 1.0):
        super().__init__(actions=actions, n_jobs=n_jobs,
                         context_agent_type=context_agent_type, available_operations=available_operations,
                         enable_logging=enable_logging, decaying_factor=decaying_factor)
        context_size = get_context_size(context_agent_type, available_operations)

        self._agent = NeuralMAB(arms=self._indices,
                                context_size=context_size,
                                n_jobs=n_jobs)



from dataclasses import dataclass

from golem.core.optimisers.optimization_parameters import GraphRequirements


@dataclass
class DynamicGraphRequirements(GraphRequirements):
    """ Class for using custom domain specific graph requirements. """
    def __init__(self, attributes: dict):
        for attribute, value in attributes.items():
            setattr(self, attribute, value)



from copy import deepcopy
from dataclasses import dataclass, field
from datetime import timedelta
from typing import List, Dict, Generic, Tuple, Any, Optional

import numpy as np
from iOpt.output_system.listeners.console_outputers import ConsoleOutputListener
from iOpt.problem import Problem
from iOpt.solver import Solver
from iOpt.solver_parametrs import SolverParameters
from iOpt.trial import Point, FunctionValue

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.optimisers.genetic.evaluation import determine_n_jobs
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import ObjectiveEvaluate
from golem.core.tuning.search_space import SearchSpace, get_node_operation_parameter_label, convert_parameters
from golem.core.tuning.tuner_interface import BaseTuner, DomainGraphForTune
from golem.utilities.data_structures import ensure_wrapped_in_sequence


@dataclass
class IOptProblemParameters:
    float_parameters_names: List[str] = field(default_factory=list)
    discrete_parameters_names: List[str] = field(default_factory=list)
    lower_bounds_of_float_parameters: List[float] = field(default_factory=list)
    upper_bounds_of_float_parameters: List[float] = field(default_factory=list)
    discrete_parameters_vals: List[List[Any]] = field(default_factory=list)

    @staticmethod
    def from_parameters_dicts(float_parameters_dict: Optional[Dict[str, List]] = None,
                              discrete_parameters_dict: Optional[Dict[str, List]] = None):
        float_parameters_dict = float_parameters_dict or {}
        discrete_parameters_dict = discrete_parameters_dict or {}

        float_parameters_names = list(float_parameters_dict.keys())
        discrete_parameters_names = list(discrete_parameters_dict.keys())

        lower_bounds_of_float_parameters = [bounds[0] for bounds in float_parameters_dict.values()]
        upper_bounds_of_float_parameters = [bounds[1] for bounds in float_parameters_dict.values()]
        discrete_parameters_vals = [values_set for values_set in discrete_parameters_dict.values()]

        return IOptProblemParameters(float_parameters_names,
                                     discrete_parameters_names,
                                     lower_bounds_of_float_parameters,
                                     upper_bounds_of_float_parameters,
                                     discrete_parameters_vals)


class GolemProblem(Problem, Generic[DomainGraphForTune]):
    def __init__(self, graph: DomainGraphForTune,
                 objective_evaluate: ObjectiveEvaluate,
                 problem_parameters: IOptProblemParameters,
                 objectives_number: int = 1):
        super().__init__()
        self.objective_evaluate = objective_evaluate
        self.graph = graph

        self.number_of_objectives = objectives_number
        self.number_of_constraints = 0

        self.discrete_variable_names = problem_parameters.discrete_parameters_names
        self.discrete_variable_values = problem_parameters.discrete_parameters_vals
        self.number_of_discrete_variables = len(self.discrete_variable_names)

        self.float_variable_names = problem_parameters.float_parameters_names
        self.lower_bound_of_float_variables = problem_parameters.lower_bounds_of_float_parameters
        self.upper_bound_of_float_variables = problem_parameters.upper_bounds_of_float_parameters
        self.number_of_float_variables = len(self.float_variable_names)

        self._default_metric_value = np.inf

    def calculate(self, point: Point, function_value: FunctionValue) -> FunctionValue:
        new_parameters = self.get_parameters_dict_from_iopt_point(point)
        BaseTuner.set_arg_graph(self.graph, new_parameters)
        graph_fitness = self.objective_evaluate(self.graph)
        metric_value = graph_fitness.value if graph_fitness.valid else self._default_metric_value
        function_value.value = metric_value
        return function_value

    def get_parameters_dict_from_iopt_point(self, point: Point) -> Dict[str, Any]:
        """Constructs a dict with all hyperparameters """
        float_parameters = dict(zip(self.float_variable_names, point.float_variables)) \
            if point.float_variables is not None else {}
        discrete_parameters = dict(zip(self.discrete_variable_names, point.discrete_variables)) \
            if point.discrete_variables is not None else {}

        parameters_dict = {**float_parameters, **discrete_parameters}
        return parameters_dict


class IOptTuner(BaseTuner):
    """
    Class for hyperparameters optimization based on iOpt library

    Args:
        objective_evaluate: objective to optimize
        adapter: the function for processing of external object that should be optimized
        iterations: max number of iterations
        search_space: SearchSpace instance
        n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's)
        eps: The accuracy of the solution of the problem. Less value - higher search accuracy, less likely to stop
            prematurely.
        r: Reliability parameter. Higher r is slower convergence, higher probability of finding a global minimum.
        evolvent_density: Density of the evolvent. By default :math:`2^{-10}` on hypercube :math:`[0,1]^N`,
             which means, that the maximum search accuracy is :math:`2^{-10}`.
        eps_r: Parameter that affects the speed of solving the task. epsR = 0 - slow convergence
             to the exact solution, epsR>0 - quick converge to the neighborhood of the solution.
        refine_solution: if true, then the solution will be refined with local search.
        deviation: required improvement (in percent) of a metric to return tuned graph.
            By default, ``deviation=0.05``, which means that tuned graph will be returned
            if it's metric will be at least 0.05% better than the initial.
    """

    def __init__(self, objective_evaluate: ObjectiveEvaluate,
                 search_space: SearchSpace,
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 iterations: int = 100,
                 timeout: timedelta = timedelta(minutes=5),
                 n_jobs: int = -1,
                 eps: float = 0.001,
                 r: float = 2.0,
                 evolvent_density: int = 10,
                 eps_r: float = 0.001,
                 refine_solution: bool = False,
                 deviation: float = 0.05, **kwargs):
        super().__init__(objective_evaluate,
                         search_space,
                         adapter,
                         iterations=iterations,
                         timeout=timeout,
                         n_jobs=n_jobs,
                         deviation=deviation, **kwargs)
        self.n_jobs = determine_n_jobs(self.n_jobs)
        self.solver_parameters = SolverParameters(r=np.double(r),
                                                  eps=np.double(eps),
                                                  iters_limit=iterations,
                                                  evolvent_density=evolvent_density,
                                                  eps_r=np.double(eps_r),
                                                  refine_solution=refine_solution,
                                                  number_of_parallel_points=self.n_jobs,
                                                  # round timeout up to avoid timeout=0 (minutes) passing to iOpt
                                                  timeout=np.ceil(timeout.total_seconds()/60) if self.timeout else -1)

    def _tune(self, graph: DomainGraphForTune, show_progress: bool = True) -> DomainGraphForTune:
        problem_parameters, initial_parameters = self._get_parameters_for_tune(graph)

        has_parameters_to_optimize = (len(problem_parameters.discrete_parameters_names) > 0 or
                                      len(problem_parameters.float_parameters_names) > 0)
        self.objectives_number = len(ensure_wrapped_in_sequence(self.init_metric))
        is_multi_objective = self.objectives_number > 1

        if self._check_if_tuning_possible(graph, has_parameters_to_optimize, supports_multi_objective=True):
            if initial_parameters:
                initial_point = Point(**initial_parameters)
                self.solver_parameters.start_point = initial_point

            problem = GolemProblem(graph, self.objective_evaluate, problem_parameters, self.objectives_number)
            solver = Solver(problem, parameters=self.solver_parameters)

            if show_progress:
                console_output = ConsoleOutputListener(mode='full')
                solver.add_listener(console_output)

            solver.solve()
            solution = solver.get_results()
            if not is_multi_objective:
                best_point = solution.best_trials[0].point
                best_parameters = problem.get_parameters_dict_from_iopt_point(best_point)
                tuned_graphs = self.set_arg_graph(graph, best_parameters)
                self.was_tuned = True
            else:
                tuned_graphs = []
                for best_trial in solution.best_trials:
                    best_parameters = problem.get_parameters_dict_from_iopt_point(best_trial.point)
                    tuned_graph = self.set_arg_graph(deepcopy(graph), best_parameters)
                    tuned_graphs.append(tuned_graph)
                    self.was_tuned = True
        else:
            tuned_graphs = graph

        return tuned_graphs

    def _get_parameters_for_tune(self, graph: OptGraph) -> Tuple[IOptProblemParameters, dict]:
        """ Method for defining the search space

        Args:
            graph: graph to be tuned

        Returns:
            parameters_dict: dict with operation names and parameters
            initial_parameters: dict with initial parameters of the graph
        """
        float_parameters_dict = {}
        discrete_parameters_dict = {}
        has_init_parameters = any(len(node.parameters) > 0 for node in graph.nodes)
        initial_parameters = {'float_variables': [], 'discrete_variables': []} if has_init_parameters else None
        for node_id, node in enumerate(graph.nodes):
            operation_name = node.name

            # Assign unique prefix for each model hyperparameter
            # label - number of node in the graph
            float_node_parameters, discrete_node_parameters = get_node_parameters_for_iopt(
                self.search_space,
                node_id,
                operation_name)
            if has_init_parameters:
                # Set initial parameters for search
                for parameter, bounds in convert_parameters(float_node_parameters).items():
                    # If parameter is not set use parameter minimum possible value
                    initial_value = node.parameters.get(parameter) or bounds[0]
                    initial_parameters['float_variables'].append(initial_value)

                for parameter, values in convert_parameters(discrete_node_parameters).items():
                    # If parameter is not set use the last value
                    initial_value = node.parameters.get(parameter) or values[-1]
                    initial_parameters['discrete_variables'].append(initial_value)

            float_parameters_dict.update(float_node_parameters)
            discrete_parameters_dict.update(discrete_node_parameters)
        parameters_dict = IOptProblemParameters.from_parameters_dicts(float_parameters_dict, discrete_parameters_dict)
        return parameters_dict, initial_parameters


def get_node_parameters_for_iopt(search_space: SearchSpace, node_id: int, operation_name: str) \
        -> Tuple[Dict[str, List], Dict[str, List]]:
    """
    Method for forming dictionary with hyperparameters of node operation for the ``IOptTuner``

    Args:
        search_space: SearchSpace with parameters per operation
        node_id: number of node in graph.nodes list
        operation_name: name of operation in the node

    Returns:
        float_parameters_dict: dictionary-like structure with labeled float hyperparameters
        and their range per operation
        discrete_parameters_dict: dictionary-like structure with labeled discrete hyperparameters
        and their range per operation
    """
    # Get available parameters for operation
    parameters_dict = search_space.parameters_per_operation.get(operation_name, {})

    discrete_parameters_dict = {}
    float_parameters_dict = {}
    categorical_parameters_dict = {}

    for parameter_name, parameter_properties in parameters_dict.items():
        node_op_parameter_name = get_node_operation_parameter_label(node_id, operation_name, parameter_name)

        parameter_type = parameter_properties.get('type')
        if parameter_type == 'discrete':
            discrete_parameters_dict.update({node_op_parameter_name: list(range(*parameter_properties
                                                                                .get('sampling-scope')))})
        elif parameter_type == 'continuous':
            float_parameters_dict.update({node_op_parameter_name: parameter_properties
                                         .get('sampling-scope')})
        elif parameter_type == 'categorical':
            categorical_parameters_dict.update({node_op_parameter_name: parameter_properties
                                               .get('sampling-scope')[0]})

    # IOpt does not distinguish between discrete and categorical parameters
    discrete_parameters_dict = {**discrete_parameters_dict, **categorical_parameters_dict}
    return float_parameters_dict, discrete_parameters_dict



import math
import os
from functools import partial
from typing import Callable, List, Union, Optional

from matplotlib import pyplot as plt, animation

from golem.core.dag.graph import Graph
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.visualisation.graph_viz import GraphVisualizer
from golem.visualisation.opt_history.history_visualization import HistoryVisualization


class GenealogicalPath(HistoryVisualization):
    def visualize(self, graph_dist: Callable[[Graph, Graph], float] = None, target_graph: Graph = None,
                  evolution_time_s: float = 8., hold_result_time_s: float = 2.,
                  save_path: Optional[Union[os.PathLike, str]] = None, show: bool = False):
        """
        Takes the best individual from the resultant generation and traces its genealogical path
        taking the most similar parent each time (or the first parent if no similarity measure is provided).
        That makes the picture more stable (and hence comprehensible) and the evolution process more apparent.

        Saves the result as a GIF with the following layout:
        - target graph (if provided) is displayed on the left,
        - evolving graphs go as the next subplot, they evolve from the first generation to the last,
        - and the fitness plot on the right shows fitness dynamics as the graphs evolve.

        :param graph_dist: a function to measure the distance between two graphs. If not provided, all graphs are
            treated as equally distant.
            Works on optimization graphs, not domain graphs. If your distance metric works on domain graphs,
            adapt it with `adapter.adapt_func(your_metric)`.
        :param target_graph: the graph to compare the genealogical path with. Again, optimization graph is expected.
            If provided, it will be displayed on the left throughout the animation.
        :param save_path: path to save the visualization (won't be saved if it's None).
            GIF of video extension is expected.
        :param show: whether to show the visualization.
        :param evolution_time_s: time in seconds for the part of the animation where the evolution process is shown.
        :param hold_result_time_s: time in seconds for the part of the animation where the final result is shown.
        """
        # Treating all graphs as equally distant if there's no reasonable way to compare them:
        graph_dist = graph_dist or (lambda g1, g2: 1)

        def draw_graph(graph: Graph, ax, title, highlight_title=False):
            ax.clear()
            ax.set_title(title, fontsize=22, color='green' if highlight_title else 'black')
            GraphVisualizer(graph).draw_nx_dag(ax, node_names_placement='legend')

        try:
            last_internal_graph = self.history.archive_history[-1][0]
            genealogical_path = trace_genealogical_path(last_internal_graph, graph_dist)
        except Exception as e:
            # At least `Individual.parents_from_prev_generation` my fail
            self.log.error(f"Failed to trace genealogical path: {e}")
            return

        figure_width = 5
        width_ratios = [1.3, 0.7]
        if target_graph is not None:
            width_ratios = [1.3] + width_ratios

        fig, axes = plt.subplots(
            1, len(width_ratios),
            figsize=(figure_width * sum(width_ratios), figure_width),
            gridspec_kw={'width_ratios': width_ratios}
        )
        evo_ax, fitness_ax = axes[-2:]
        if target_graph is not None:
            draw_graph(target_graph, axes[0], "Target graph")  # Persists throughout the animation

        fitnesses_along_path = list(map(lambda ind: ind.fitness.value, genealogical_path))
        generations_along_path = list(map(lambda ind: ind.native_generation, genealogical_path))

        def render_frame(frame_index):
            path_index = min(frame_index, len(genealogical_path) - 1)
            is_hold_stage = frame_index >= len(genealogical_path)

            draw_graph(
                genealogical_path[path_index].graph, evo_ax,
                f"Evolution process,\ngeneration {generations_along_path[path_index]}/{generations_along_path[-1]}",
                highlight_title=is_hold_stage
            )
            # Select only the genealogical path
            fitness_ax.clear()
            plot_fitness_with_axvline(
                generations=generations_along_path,
                fitnesses=fitnesses_along_path,
                ax=fitness_ax,
                axvline_x=generations_along_path[path_index],
                current_fitness=fitnesses_along_path[path_index]
            )
            return evo_ax, fitness_ax

        frames = len(genealogical_path) + int(
            math.ceil(len(genealogical_path) * hold_result_time_s / (hold_result_time_s + evolution_time_s))
        )
        seconds_per_frame = (evolution_time_s + hold_result_time_s) / frames
        fps = math.ceil(1 / seconds_per_frame)

        anim = animation.FuncAnimation(fig, render_frame, repeat=False, frames=frames,
                                       interval=1000 * seconds_per_frame)

        try:
            if save_path is not None:
                anim.save(save_path, fps=fps)
            if show:
                plt.show()
        except Exception as e:
            self.log.error(f"Failed to render the genealogical path: {e}")


def trace_genealogical_path(individual: Individual, graph_dist: Callable[[Graph, Graph], float]) -> List[Individual]:
    # Choose nearest parent each time:
    genealogical_path: List[Individual] = [individual]
    while genealogical_path[-1].parents_from_prev_generation:
        genealogical_path.append(max(
            genealogical_path[-1].parents_from_prev_generation,
            key=partial(graph_dist, genealogical_path[-1])
        ))

    return list(reversed(genealogical_path))


def plot_fitness_with_axvline(generations: List[int], fitnesses: List[float], ax: plt.Axes, current_fitness: float,
                              axvline_x: int = None):
    ax.plot(generations, fitnesses)
    ax.set_title(f'Fitness dynamic,\ncurrent: {current_fitness}', fontsize=22)
    ax.set_xlabel('Generation')
    ax.set_ylabel('Fitness')
    if axvline_x is not None:
        ax.axvline(x=axvline_x, color='black')
    return ax



from functools import partial
from typing import Callable, Union


def get_callable_name(action: Union[Callable, str]) -> str:
    if isinstance(action, str):
        return action
    if isinstance(action, partial):
        return action.func.__name__
    try:
        return action.__name__
    except AttributeError:
        return str(action)



import copy
import math
from typing import List, Any, Union, Dict

from golem.utilities.requirements_notificator import warn_requirement

try:
    import torch
except ModuleNotFoundError:
    warn_requirement('torch', 'other_requirements/adaptive.txt')

import numpy as np
from mabwiser.mab import MAB, LearningPolicy, NeighborhoodPolicy
from mabwiser.utils import Arm, Constants, Num

from golem.core.log import default_log

import warnings

warnings.filterwarnings("ignore")


class NeuralMAB(MAB):
    """
    Neural Multi-Armed Bandit.
    The main concept is explained in the article: https://arxiv.org/abs/2012.01780.
    Deep representation is formed with NN and Contextual Multi-Armed Bandit is integrated to choose arm.
    """

    def __init__(self, arms: List[Arm],
                 learning_policy: Any = LearningPolicy.UCB1(alpha=1.25),
                 neighborhood_policy: Any = NeighborhoodPolicy.Clusters(),
                 seed: int = Constants.default_seed,
                 context_size: int = 1,
                 n_jobs: int = 1):

        super().__init__(arms, learning_policy, neighborhood_policy, seed, n_jobs)
        self.nn_with_se = NNWithShallowExploration(context_size=context_size, arms_count=len(arms))
        self.arms = arms
        self.seed = seed
        self.n_jobs = n_jobs
        self.log = default_log('NeuralMAB')
        # to track when GNN needs to be updated
        self.iter = 0
        self._indices = list(range(len(arms)))
        self._mab = MAB(arms=self._indices,
                        learning_policy=learning_policy,
                        neighborhood_policy=neighborhood_policy,
                        n_jobs=n_jobs)
        self.is_fitted = False

    def _initial_fit_mab(self, context: Any):
        """ Initial fit for Contextual Multi-Armed Bandit.
        At this step, all hands are assigned the same weights with the very first context
        that is fed to the bandit. """
        # initial fit for mab
        n = len(self.arms)
        uniform_rewards = [1. / n] * n
        deep_context = self._get_deep_context(context=context)
        self._mab.fit(decisions=self._indices, rewards=uniform_rewards, contexts=n * [deep_context])
        self.is_fitted = True

    def partial_fit(self, decisions: List[Any], rewards: List[float], contexts: List[Any] = None):

        # get deep contexts, calculate regret and update weights for NN (once in _H_q iters)
        deep_contexts = self.nn_with_se.partial_fit(iter=self.iter, decisions=decisions,
                                                    rewards=rewards, contexts=contexts)
        self.iter += 1

        # update contextual mab with deep contexts
        self._mab.partial_fit(decisions=decisions, contexts=deep_contexts, rewards=rewards)

    def predict(self, contexts: Any = None) -> Union[Arm, List[Arm]]:
        """ Predicts which arm to pull to get maximum reward. """
        if not self.is_fitted:
            self._initial_fit_mab(context=contexts)
        deep_context = self._get_deep_context(context=contexts)
        a_choose = self._mab.predict(contexts=[deep_context])
        return a_choose

    def predict_expectations(self, contexts: Any = None) -> Union[Dict[Arm, Num], List[Dict[Arm, Num]]]:
        """ Returns expected reward for each arm. """
        if not self.is_fitted:
            self._initial_fit_mab(context=contexts)
        deep_context = self._get_deep_context(context=contexts)
        return self._mab.predict_expectations(contexts=[deep_context])

    def _get_deep_context(self, context: Any) -> List[Any]:
        """ Returns deep representation of context. """
        temp = self.nn_with_se.transfer(context, 0, len(self.arms))
        feat = self.nn_with_se.feature_extractor(temp, self.nn_with_se.W).numpy().squeeze()
        return list(feat)


class NNWithShallowExploration:
    """ Neural Network with shallow exploration which means that weights are updated every _H_q iterations. """

    def __init__(self, context_size: int, arms_count: int):
        """
        Initial fit for NN.
        beta -- parameter for UCB exploration
        H_q -- how many time steps to update NN
        interT -- internal steps for GD
        """
        self._beta = 0.02
        self._lambd = 1
        self._lr = 0.001
        self._H_q = 5
        self._interT = 1000
        self._hidden_dim = [1000, 1000]
        hid_dim_lst = self._hidden_dim
        dim_second_last = self._hidden_dim[-1] * 2

        dim_for_init = [context_size + arms_count] + hid_dim_lst + [1]
        self.arms_count = arms_count
        self.W0, total_dim = self._initialization(dim_for_init)
        self.LAMBDA = self._lambd * torch.eye(dim_second_last, dtype=torch.double)
        self.bb = torch.zeros(self.LAMBDA.size()[0], dtype=torch.double).view(-1, 1)

        theta = np.random.randn(dim_second_last, 1) / np.sqrt(dim_second_last)
        self.theta = torch.from_numpy(theta)

        self.THETA_action = torch.tensor([])
        self.CONTEXT_action = torch.tensor([])
        self.REWARD_action = torch.tensor([])
        self.result_neuralucb = []
        self.W = copy.deepcopy(self.W0)
        self.summ = 0
        self.log = default_log('NNWithShallowExploration')

    def partial_fit(self, iter: int,
                    decisions: List[Any], rewards: List[float], contexts: List[Any] = None):
        deep_contexts = []

        # update NN and calculate reward
        for decision, context, reward in zip(decisions, contexts, rewards):

            # calculate reward
            temp = self.transfer(context, decision, self.arms_count)
            feat = self.feature_extractor(temp, self.W)
            deep_contexts.append(list(feat.numpy().squeeze()))
            expected_reward = torch.mm(self.theta.view(1, -1), feat) + self._beta * self.UCB(self.LAMBDA, feat)

            self.summ += (expected_reward - reward)
            self.result_neuralucb.append(self.summ)

            # gather dataset for next NN training (context_action and reward_action)
            if np.mod(iter, self._H_q) == 0:
                context_action = temp
                reward_action = torch.tensor([reward], dtype=torch.double)
            else:
                context_action = torch.cat((self.CONTEXT_action, temp), 1)
                reward_action = torch.cat((self.REWARD_action, torch.tensor([reward], dtype=torch.double)), 0)

            # update LAMBDA and bb
            self.LAMBDA += torch.mm(self.feature_extractor(temp, self.W),
                                    self.feature_extractor(temp, self.W).t())
            self.bb += reward * self.feature_extractor(temp, self.W)
            theta = torch.linalg.solve(self.LAMBDA, self.bb)

            if np.mod(iter, self._H_q) == 0:
                theta_action = theta.view(-1, 1)
            else:
                theta_action = torch.cat((self.THETA_action, theta.view(-1, 1)), 1)

        # update weight of NN
        if np.mod(iter, self._H_q) == self._H_q - 1:
            self.log.info(f'Current regret: {self.summ}')
            self.W = self.train_with_shallow_exploration(context_action, reward_action, self.W0,
                                                         self._interT, self._lr, theta_action, self._H_q)
        return deep_contexts

    @staticmethod
    def UCB(A, phi):
        """ Ucb term. """
        try:
            tmp = torch.linalg.solve(A, phi)
        except Exception:
            tmp = torch.Tensor(np.linalg.solve(A, phi))

        return torch.sqrt(torch.mm(torch.transpose(phi, 0, 1).double(), tmp.double()))

    @staticmethod
    def transfer(c, a, arm_size):
        """
        Transfer an array context + action to new context with dimension 2*(__context__ + __armsize__).
        """
        action = np.zeros(arm_size)
        action[a] = 1
        c_final = np.append(c, action)
        c_final = torch.from_numpy(c_final)
        c_final = c_final.view((len(c_final), 1))
        c_final = c_final.repeat(2, 1)
        return c_final

    def train_with_shallow_exploration(self, X, Y, W_start, T, et, THETA, H):
        """ Gd-based model training with shallow exploration
        Dataset X, label Y. """
        W = copy.deepcopy(W_start)
        X = X[:, -H:]
        Y = Y[-H:]
        THETA = THETA[:, -H:]

        prev_loss = 1000000
        prev_loss_1k = 1000000
        for i in range(0, T):
            grad = self._gradient_loss(X, Y, W, THETA)
            for j in range(0, len(W) - 1):
                W[j] = W[j] - et * grad[j]

            curr_loss = self._loss(X, Y, W, THETA)
            if i % 100 == 0:
                print('------', curr_loss)
                if curr_loss > prev_loss_1k:
                    et = et * 0.1
                    print('lr/10 to', et)

                prev_loss_1k = curr_loss

            # early stopping
            if abs(curr_loss - prev_loss) < 1e-6:
                break
            prev_loss = curr_loss
        return W

    @staticmethod
    def _initialization(dim):
        """ Initialization.
        dim consists of (d1, d2,...), where dl = 1 (placeholder, deprecated). """
        w = []
        total_dim = 0
        for i in range(0, len(dim) - 1):
            if i < len(dim) - 2:
                temp = np.random.randn(dim[i + 1], dim[i]) / np.sqrt(dim[i + 1])
                temp = np.kron(np.eye(2, dtype=int), temp)
                temp = torch.from_numpy(temp)
                w.append(temp)
                total_dim += dim[i + 1] * dim[i] * 4
            else:
                temp = np.random.randn(dim[i + 1], dim[i]) / np.sqrt(dim[i])
                temp = np.kron([[1, -1]], temp)
                temp = torch.from_numpy(temp)
                w.append(temp)
                total_dim += dim[i + 1] * dim[i] * 2

        return w, total_dim

    @staticmethod
    def feature_extractor(x, W):
        """ Functions feature extractor.
        x is the input, dimension is d; W is the list of parameter matrices. """
        depth = len(W)
        output = x
        for i in range(0, depth - 1):
            output = torch.mm(W[i], output)
            output = (output - output.mean()) / (output.std() + 1e-8)
            output = output.clamp(min=0)

        output = output * math.sqrt(W[depth - 1].size()[1])
        return output

    def _gradient_loss(self, X, Y, W, THETA):
        """ Return a list of grad, satisfying that W[i] = W[i] - grad[i] for single context x. """
        depth = len(W)
        num_sample = Y.shape[0]
        loss = []
        grad = []
        relu = []
        output = X
        loss.append(output)
        for i in range(0, depth - 1):
            output = torch.mm(W[i], output)
            relu.append(output)
            output = output.clamp(min=0)
            loss.append(output)

        THETA_t = torch.transpose(THETA, 0, 1).view(num_sample, 1, -1)
        output_t = torch.transpose(output, 0, 1).view(num_sample, -1, 1)
        output = torch.bmm(THETA_t, output_t).squeeze().view(1, -1)

        loss.append(output)

        feat = self.feature_extractor(X, W)
        feat_t = torch.transpose(feat, 0, 1).view(num_sample, -1, 1)
        output_t = torch.bmm(THETA_t, feat_t).squeeze().view(1, -1)

        # backward gradient propagation
        back = output_t - Y
        back = back.double()
        grad_t = torch.mm(back, loss[depth - 1].t())
        grad.append(grad_t)

        for i in range(1, depth):
            back = torch.mm(W[depth - i].t(), back)
            back[relu[depth - i - 1] < 0] = 0
            grad_t = torch.mm(back, loss[depth - i - 1].t())
            grad.append(grad_t)

        grad1 = []
        for i in range(0, depth):
            grad1.append(grad[depth - 1 - i] * math.sqrt(W[depth - 1].size()[1]) / len(X[0, :]))

        return grad1

    def _loss(self, X, Y, W, THETA):
        # total loss
        num_sample = len(X[0, :])
        output = self.feature_extractor(X, W)
        THETA_t = torch.transpose(THETA, 0, 1).view(num_sample, 1, -1)
        output_t = torch.transpose(output, 0, 1).view(num_sample, -1, 1)
        output_y = torch.bmm(THETA_t, output_t).squeeze().view(1, -1)

        summ = (Y - output_y).pow(2).sum() / num_sample
        return summ



from typing import Optional

import numpy as np
from joblib import cpu_count

import random
from golem.utilities.random import RandomStateHandler


def determine_n_jobs(n_jobs=-1, logger=None):
    cpu_num = cpu_count()
    if n_jobs > cpu_num:
        n_jobs = cpu_num
    elif n_jobs <= 0:
        if n_jobs <= -cpu_num - 1 or n_jobs == 0:
            raise ValueError(f"Unproper `n_jobs` = {n_jobs}. "
                             f"`n_jobs` should be between ({-cpu_num}, {cpu_num}) except 0")
        n_jobs = cpu_num + 1 + n_jobs
    if logger:
        logger.info(f"Number of used CPU's: {n_jobs}")
    return n_jobs


def set_random_seed(seed: Optional[int]):
    """ Sets random seed for evaluation of models. """
    if seed is not None:
        np.random.seed(seed)
        random.seed(seed)
        RandomStateHandler.MODEL_FITTING_SEED = seed



import logging
from typing import Optional

from golem.api.api_utils.api_params import ApiParams
from golem.core.constants import DEFAULT_API_TIMEOUT_MINUTES
from golem.core.log import Log, default_log
from golem.utilities.utilities import set_random_seed


class GOLEM:
    """
    Main class for GOLEM API.

    Args:
        :param timeout: timeout for optimization.
        :param seed: value for a fixed random seed.
        :param logging_level: logging levels are the same as in `logging <https://docs.python.org/3/library/logging.html>`_.

            .. details:: Possible options:

                - ``50`` -> critical
                - ``40`` -> error
                - ``30`` -> warning
                - ``20`` -> info
                - ``10`` -> debug
                - ``0`` -> nonset
        :param n_jobs: num of ``n_jobs`` for parallelization (set to ``-1`` to use all cpu's). Defaults to ``-1``.
        :param graph_requirements_class: class to specify custom graph requirements.
        Must be inherited from GraphRequirements class.

        :param crossover_prob: crossover probability (chance that two individuals will be mated).

        ``GPAlgorithmParameters`` parameters
        :param mutation_prob: mutation probability (chance that an individual will be mutated).
        :param variable_mutation_num: flag to apply mutation one or few times for individual in each iteration.
        :param max_num_of_operator_attempts: max number of unsuccessful evo operator attempts before continuing.
        :param mutation_strength: strength of mutation in tree (using in certain mutation types)
        :param min_pop_size_with_elitism: minimal population size with which elitism is applicable
        :param required_valid_ratio: ratio of valid individuals on next population to continue optimization.

        Used in `ReproductionController` to compensate for invalid individuals. See the class for details.

        :param adaptive_mutation_type: enables adaptive Mutation agent.
        :param context_agent_type: enables graph encoding for Mutation agent.

        Adaptive mutation agent uses specified algorithm. 'random' type is the default non-adaptive version.
        Requires crossover_types to be CrossoverTypesEnum.none for correct adaptive learning,
        so that fitness changes depend only on agent's actions (chosen mutations).
        ``MutationAgentTypeEnum.bandit`` uses Multi-Armed Bandit (MAB) learning algorithm.
        ``MutationAgentTypeEnum.contextual_bandit`` uses contextual MAB learning algorithm.
        ``MutationAgentTypeEnum.neural_bandit`` uses contextual MAB learning algorithm with Deep Neural encoding.

        Parameter `context_agent_type` specifies implementation of graph/node encoder for adaptive
        mutation agent. It is relevant for contextual and neural bandits.

        :param decaying_factor: decaying factor for Multi-Armed Bandits for managing the profit from operators
        The smaller the value of decaying_factor, the larger the influence for the best operator.
        :param window_size: the size of sliding window for Multi-Armed Bandits to decrease variance.
        The window size is measured by the number of individuals to consider.


        :param selection_types: Sequence of selection operators types
        :param crossover_types: Sequence of crossover operators types
        :param mutation_types: Sequence of mutation operators types
        :param elitism_type: type of elitism operator evolution

        :param regularization_type: type of regularization operator

        Regularization attempts to cut off the subtrees of the graph. If the truncated graph
        is not worse than the original, then it enters the new generation as a simpler solution.
        Regularization is not used by default, it must be explicitly enabled.

        :param genetic_scheme_type: type of genetic evolutionary scheme

        The `generational` scheme is a standard scheme of the evolutionary algorithm.
        It specifies that at each iteration the entire generation is updated.

        In the `steady_state` individuals from previous populations are mixed with the ones from new population.
        UUIDs of individuals do not repeat within one population.

        The `parameter_free` scheme is same as `steady_state` for now.

        ``GraphGenerationParams`` parameters
        :param adapter: instance of domain graph adapter for adaptation
         between domain and optimization graphs
        :param rules_for_constraint: collection of constraints for graph verification
        :param advisor: instance providing task and context-specific advices for graph changes
        :param node_factory: instance for generating new nodes in the process of graph search
        :param remote_evaluator: instance of delegate evaluator for evaluation of graphs

        ``GraphRequirements`` parameters
        :param start_depth: start value of adaptive tree depth
        :param max_depth: max depth of the resulting graph
        :param min_arity: min number of parents for node
        :param max_arity: max number of parents for node

        Also, custom domain specific parameters can be specified here. These parameters can be then used in
        ``DynamicGraphRequirements`` as fields.
    """
    def __init__(self,
                 timeout: Optional[float] = DEFAULT_API_TIMEOUT_MINUTES,
                 seed: Optional[int] = None,
                 logging_level: int = logging.INFO,
                 n_jobs: int = -1,
                 **all_parameters):
        set_random_seed(seed)
        self.log = self._init_logger(logging_level)

        self.api_params = ApiParams(input_params=all_parameters,
                                    n_jobs=n_jobs,
                                    timeout=timeout)
        self.gp_algorithm_parameters = self.api_params.get_gp_algorithm_parameters()
        self.graph_generation_parameters = self.api_params.get_graph_generation_parameters()
        self.graph_requirements = self.api_params.get_graph_requirements()

    def optimise(self, **custom_optimiser_parameters):
        """ Method to start optimisation process.
        `custom_optimiser_parameters` parameters can be specified additionally to use it directly in optimiser.
        """
        common_params = self.api_params.get_actual_common_params()
        optimizer_cls = common_params['optimizer']
        objective = common_params['objective']
        initial_graphs = common_params['initial_graphs']

        self.optimiser = optimizer_cls(objective,
                                       initial_graphs,
                                       self.graph_requirements,
                                       self.graph_generation_parameters,
                                       self.gp_algorithm_parameters,
                                       **custom_optimiser_parameters)

        found_graphs = self.optimiser.optimise(objective)
        return found_graphs

    @staticmethod
    def _init_logger(logging_level: int):
        # reset logging level for Singleton
        Log().reset_logging_level(logging_level)
        return default_log(prefix='GOLEM logger')



from datetime import timedelta
from functools import partial
from typing import Callable, Optional

from hyperopt import tpe, fmin, space_eval, Trials

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.constants import MIN_TIME_FOR_TUNING_IN_SEC
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import ObjectiveFunction
from golem.core.tuning.hyperopt_tuner import HyperoptTuner, get_node_parameters_for_hyperopt
from golem.core.tuning.search_space import SearchSpace
from golem.core.tuning.tuner_interface import DomainGraphForTune


class SequentialTuner(HyperoptTuner):
    """
    Class for hyperparameters optimization for all nodes sequentially
    """

    def __init__(self, objective_evaluate: ObjectiveFunction,
                 search_space: SearchSpace,
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 iterations: int = 100,
                 early_stopping_rounds: Optional[int] = None,
                 timeout: timedelta = timedelta(minutes=5),
                 n_jobs: int = -1,
                 deviation: float = 0.05,
                 algo: Callable = tpe.suggest,
                 inverse_node_order: bool = False, **kwargs):
        super().__init__(objective_evaluate,
                         search_space,
                         adapter,
                         iterations,
                         early_stopping_rounds, timeout,
                         n_jobs,
                         deviation,
                         algo, **kwargs)

        self.inverse_node_order = inverse_node_order

    def _tune(self, graph: DomainGraphForTune, **kwargs) -> DomainGraphForTune:
        """ Method for hyperparameters tuning on the entire graph

        Args:
            graph: graph which hyperparameters will be tuned
        """
        remaining_time = self._get_remaining_time()
        if self._check_if_tuning_possible(graph, parameters_to_optimize=True, remaining_time=remaining_time):
            # Calculate amount of iterations we can apply per node
            nodes_amount = graph.length
            iterations_per_node = round(self.iterations / nodes_amount)
            iterations_per_node = int(iterations_per_node)
            if iterations_per_node == 0:
                iterations_per_node = 1

            # Calculate amount of seconds we can apply per node
            if remaining_time is not None:
                seconds_per_node = round(remaining_time / nodes_amount)
                seconds_per_node = int(seconds_per_node)
            else:
                seconds_per_node = None

            # Tuning performed sequentially for every node - so get ids of nodes
            nodes_ids = self.get_nodes_order(nodes_number=nodes_amount)
            for node_id in nodes_ids:
                node = graph.nodes[node_id]

                # Get node's parameters to optimize
                node_params, init_params = get_node_parameters_for_hyperopt(self.search_space, node_id, node)
                if not node_params:
                    self.log.info(f'"{node.name}" operation has no parameters to optimize')
                else:
                    # Apply tuning for current node
                    graph, _ = self._optimize_node(node_id=node_id,
                                                   graph=graph,
                                                   node_params=node_params,
                                                   init_params=init_params,
                                                   iterations_per_node=iterations_per_node,
                                                   seconds_per_node=seconds_per_node)
            self.was_tuned = True
        return graph

    def get_nodes_order(self, nodes_number: int) -> range:
        """ Method returns list with indices of nodes in the graph

        Args:
            nodes_number: number of nodes to get
        """

        if self.inverse_node_order is True:
            # From source data to output
            nodes_ids = range(nodes_number - 1, -1, -1)
        else:
            # From output to source data
            nodes_ids = range(0, nodes_number)

        return nodes_ids

    def tune_node(self, graph: DomainGraphForTune, node_index: int) -> DomainGraphForTune:
        """ Method for hyperparameters tuning for particular node

        Args:
            graph: graph which contains a node to be tuned
            node_index: Index of the node to tune

        Returns:
            Graph with tuned parameters in node with specified index
        """
        graph = self.adapter.adapt(graph)

        with self.timer:
            self.init_check(graph)

            node = graph.nodes[node_index]

            # Get node's parameters to optimize
            node_params, init_params = get_node_parameters_for_hyperopt(self.search_space,
                                                                        node_id=node_index,
                                                                        node=node)

            remaining_time = self._get_remaining_time()
            if self._check_if_tuning_possible(graph, len(node_params) > 1, remaining_time):
                # Apply tuning for current node
                graph, _ = self._optimize_node(graph=graph,
                                               node_id=node_index,
                                               node_params=node_params,
                                               init_params=init_params,
                                               iterations_per_node=self.iterations,
                                               seconds_per_node=remaining_time
                                               )

                self.was_tuned = True

                # Validation is the optimization do well
                final_graph = self.final_check(graph)
            else:
                final_graph = graph
                self.obtained_metric = self.init_metric
        final_graph = self.adapter.restore(final_graph)
        return final_graph

    def _optimize_node(self, graph: OptGraph,
                       node_id: int,
                       node_params: dict,
                       init_params: dict,
                       iterations_per_node: int,
                       seconds_per_node: float) -> OptGraph:
        """
        Method for node optimization

        Args:
            graph: Graph which node is optimized
            node_id: id of the current node in the graph
            node_params: dictionary with parameters for node
            iterations_per_node: amount of iterations to produce
            seconds_per_node: amount of seconds to produce

        Returns:
            updated graph with tuned parameters in particular node
        """
        remaining_time = self._get_remaining_time()
        trials = Trials()
        trials, init_trials_num = self._search_near_initial_parameters(partial(self._objective,
                                                                               graph=graph,
                                                                               node_id=node_id,
                                                                               unchangeable_parameters=init_params),
                                                                       node_params,
                                                                       init_params,
                                                                       trials,
                                                                       remaining_time)

        remaining_time = self._get_remaining_time()
        if remaining_time > MIN_TIME_FOR_TUNING_IN_SEC:
            fmin(partial(self._objective, graph=graph, node_id=node_id),
                 node_params,
                 trials=trials,
                 algo=self.algo,
                 max_evals=iterations_per_node,
                 early_stop_fn=self.early_stop_fn,
                 timeout=seconds_per_node)

        best_params = space_eval(space=node_params, hp_assignment=trials.argmin)
        is_best_trial_with_init_params = trials.best_trial.get('tid') in range(init_trials_num)
        if is_best_trial_with_init_params:
            best_params = {**best_params, **init_params}
        # Set best params for this node in the graph
        graph = self.set_arg_node(graph=graph, node_id=node_id, node_params=best_params)
        return graph, trials.best_trial['result']['loss']

    def _objective(self,
                   node_params: dict,
                   graph: OptGraph,
                   node_id: int,
                   unchangeable_parameters: Optional[dict] = None) -> float:
        """ Objective function for minimization problem

        Args:
            node_params: dictionary with parameters for node
            graph: graph to evaluate
            node_id: id of the node to which parameters should be assigned

        Returns:
            value of objective function
        """
        # replace new parameters with parameters
        if unchangeable_parameters:
            node_params = {**node_params, **unchangeable_parameters}

        # Set hyperparameters for node
        graph = self.set_arg_node(graph=graph, node_id=node_id, node_params=node_params)

        metric_value = self.get_metric_value(graph=graph)
        return metric_value
